#!/usr/bin/env python3
"""Collects logs and timings relevant to boot time performance."""

from typing import Tuple
from datetime import datetime
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from time import sleep
import subprocess
import logging
import sys
import re
import ast
import json
import socket
import platform
import io
import os
import dbus
from systemd import journal

try:
    import boot_timings.chart as chart
except ImportError:
    chart = None


def cntvct_offset():
    """Calculate the processor clock offset from the cntvct log value."""
    try:
        j = journal.Reader()
        j.this_boot()
        j.add_match(UNIT="basic.target")
        j.seek_tail()
        basic_target_timestamp = j.get_previous()["__MONOTONIC_TIMESTAMP"][
            0
        ].total_seconds()

        j.flush_matches()
        j.add_match(_SYSTEMD_UNIT="cntvct@basic.service")
        basic_target_cntvct = float(j.get_next()["MESSAGE"].strip("s."))

        # Derive the kernel clock to wall clock offset in microseconds
        return round((basic_target_cntvct - basic_target_timestamp) * 1000000)

    except KeyError:
        return 0


def directory_path(string):
    """Confirms a given path is a directory"""
    if os.path.isdir(string):
        return string
    raise NotADirectoryError(string)


def parse_cli():
    """Parse the CLI arguments and return the argparse dict."""
    parser = ArgumentParser(
        description="Collects logs and timings relevant to boot time performance.",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-o",
        "--output-directory",
        type=directory_path,
        default=".",
        help="Directory where output files will be stored",
    )
    parser.add_argument(
        "-d",
        "--description",
        type=str,
        help="Free form description to be added to the metadata",
    )
    parser.add_argument(
        "--kpi-re-pattern",
        type=str,
        default="multi-user.target|sysinit.target|Load Kernel Modules|mounted",
        help="Regular expression to match log messages on for KPI highlighting",
    )
    parser.add_argument(
        "--max-time",
        type=str,
        default="",
        help=(
            "Time in ms to limit the collection. Default to completion of switchroot. "
            "0 to not limit collection."
        ),
    )
    parser.add_argument(
        "--offset",
        type=float,
        default="0",
        help=(
            "Time in ms to offset SystemInit to measure wall clock time. Default to 0."
        ),
    )
    parser.add_argument(
        "--post-process",
        type=str,
        help=(
            "Post-process an existing '-boot_time_logs.json' from the boot_timings "
            "script. Other options except '-o' will be ignored and all information "
            "will be read from the provided file. The summary, log, and chart files "
            "will be created, and the report will be displayed."
        ),
    )

    try:
        return parser.parse_args()
    except NotADirectoryError:
        print("Target output directory path does not exist\n")
        sys.exit(1)


def run_command(command: str, timeout: int = 60) -> Tuple[int, str, str]:
    """Run a shell command and return a Tuple of the return code, stdout, and stderr"""
    # Split the command string to a list
    command_list = command.split()

    try:
        with subprocess.Popen(
            command_list,
            text=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as process:
            stdout, stderr = process.communicate(timeout=timeout)
    except subprocess.CalledProcessError as e:
        logging.error("Shell command failed: %e", e)
        sys.exit(1)
    except subprocess.TimeoutExpired as e:
        logging.error("Shell command timed out: %e", e)
        sys.exit(1)

    exit_status = process.returncode

    return (exit_status, stdout, stderr)


def get_time_diff(props, label, parent, start_key, finish_key, offset):
    """Calculate and return the time difference of start and finish keys"""
    if start_key not in props:
        return {}

    finish = props[finish_key] + offset
    start = props[start_key] + offset
    diff = finish - start

    props_dict = {
        "source": "systemd-dbus",
        "name": label,
        "activating": start,
        "activated": finish,
        "time": diff,
    }

    if parent:
        props_dict["parent"] = parent

    return props_dict


def get_time_diff_for_key_prefix(props, key_prefix, parent, offset):
    """Derive start and finish keys for logged items that use prefixes"""
    start_key = f"{key_prefix}StartTimestampMonotonic"
    finish_key = f"{key_prefix}FinishTimestampMonotonic"
    props_dict = get_time_diff(props, key_prefix, parent, start_key, finish_key, offset)
    return props_dict


def log_filter(log_item: dict, max_start_time, kpi_re_pattern: str, offset):
    """Apply filters to the log structures"""
    if (
        max_start_time == 0
        or log_item["activating"] < max_start_time
        or re.search(kpi_re_pattern, log_item["name"])
        or log_item["source"] == "systemd-dbus"
    ):
        if re.search(kpi_re_pattern, log_item["name"]):
            log_item["kpi"] = True
        # Apply the wall clock time offset
        if "activating" in log_item.keys():
            log_item["activating"] = log_item["activating"] + offset
        if "activated" in log_item.keys():
            log_item["activated"] = log_item["activated"] + offset
        if "deactivating" in log_item.keys():
            log_item["deactivating"] = log_item["deactivating"] + offset
        if "deactivated" in log_item.keys():
            log_item["deactivated"] = log_item["deactivated"] + offset
        return log_item
    # DEBUG
    # print(f"filtered {log_item}")
    # sys.exit()
    return None


def print_systemd_stages(input_list: list):
    """Print the systemd stages and timings hierarchy in a human-readable format"""
    output_dict = {}
    for item in input_list:
        if "count" in item.keys():
            output_dict[item["name"]] = {"count": item["count"]}
            continue
        output_dict[item["name"]] = {"duration": item["time"]}
        if "parent" in item.keys():
            output_dict[item["name"]]["parent"] = item["parent"]

    try:
        console_width, _ = os.get_terminal_size()
    except OSError:
        console_width = 100
    timeline_max_length = console_width - 6
    overall_duration = output_dict["Overall"]["duration"]

    # Calculate the portion lengths for the ASCII timeline
    sysinit_length = round(
        (
            (output_dict["SystemInit"]["duration"] / overall_duration)
            * timeline_max_length
        )
        - 1
    )
    kernel_length = round(
        ((output_dict["Kernel"]["duration"] / overall_duration) * timeline_max_length)
        - 1
    )
    if "InitRD" in output_dict.keys():
        initrd_length = round(
            ((output_dict["InitRD"]["duration"] / overall_duration) * timeline_max_length)
            - 1
        )
    else:
        initrd_length = 0

    switchroot_length = round(
        (
            (output_dict["Switchroot"]["duration"] / overall_duration)
            * timeline_max_length
        )
        - 1
    )

    # Set the label strings for the ASCII timeline
    sysinit_string = (
        f"SystemInit: {round(output_dict['SystemInit']['duration'] / 1000)} ms  "
    )
    kernel_string = f"Kernel: {round(output_dict['Kernel']['duration'] / 1000)} ms  "
    if "InitRD" in output_dict.keys():
        initrd_string = f"InitRD: {round(output_dict['InitRD']['duration'] / 1000)} ms  "
    else:
        initrd_string = ""
    switchroot_string = (
        f"Switchroot: {round(output_dict['Switchroot']['duration'] / 1000)} ms  "
    )

    # Print the labels with appropriate spacing
    formatted_string = "## Timeline of Boot Stage Durations\n"

    # Print the ASCII timeline
    formatted_string += (
        "["
        + "=" * sysinit_length
        + "/"
        + "=" * kernel_length
        + "/"
    )
    if "InitRD" in output_dict.keys():
        formatted_string += (
            "=" * initrd_length
            + "/"
        )
    formatted_string += (
        "=" * switchroot_length
        + "/=>\n\n"
    )

    formatted_string += "^ " + sysinit_string
    sysinit_gap = sysinit_length - len(sysinit_string) - 2
    if sysinit_gap < 1:
        formatted_string += "\n" + " " * (sysinit_length + 2)
    else:
        formatted_string += " " * (sysinit_gap + 1)
    formatted_string += "^ " + kernel_string
    kernel_gap = kernel_length - len(kernel_string) - 2
    if kernel_gap < 1:
        formatted_string += "\n" + " " * (sysinit_length + kernel_length + 2)
    else:
        formatted_string += " " * (kernel_gap + 1)
    if "InitRD" in output_dict.keys():
        formatted_string += "^ " + initrd_string
        initrd_gap = initrd_length - len(initrd_string)
        if initrd_gap < 1:
            formatted_string += "\n" + " " * (
                sysinit_length + kernel_length + initrd_length + 3
            )
        else:
            formatted_string += " " * ((initrd_length - len(initrd_string)) - 1)
    else:
        formatted_string += " "

    formatted_string += "^ " + switchroot_string + "\n"

    formatted_string += (
        f"\n## Durations (ms)\n{'Stage':<20}{'Components':<20}\t{'Start':<20}\tEnd\n"
    )
    for stage in input_list:
        if "activated" in stage.keys():
            if "parent" in stage.keys():
                formatted_string += f"{'':<20}"
                formatted_string += f"{stage['name']:<20}"
            else:
                formatted_string += f"{stage['name']:<20}"
                formatted_string += f"{'':<20}"
            formatted_string += f"\t{stage['activating'] / 1000:<20}"
            formatted_string += f"\t{stage['activated'] / 1000}\n"

    formatted_string += "\n\n## Summary\n" + organize_hierarchy(output_dict) + "\n"

    return formatted_string


def organize_hierarchy(input_dict, parent_key=None, indent=0):
    """Recursive function to create the hierarchy from the parent keys"""
    hierarchy_str = ""
    for key, value in input_dict.items():
        if value.get("parent") == parent_key:
            if value.get("count"):
                metric = f"{str(value.get('count', ''))}\n"
            elif "Service" in key:
                metric = "\n"
            else:
                metric = f"{str(value.get('duration', 0) / 1000)} ms\n"
            hierarchy_str += " " * indent + key
            hierarchy_str += ": " if ("Service" not in key) else ""
            hierarchy_str += metric
            children_str = organize_hierarchy(input_dict, key, indent + 2)
            hierarchy_str += children_str
    return hierarchy_str


def kpi_table(logs):
    """Returns a tab-separated table of KPIs and timestamps"""
    table = f"## KPI Timestamps (ms)\n{'KPI':<50}\tTimestamp\n"
    for log in logs:
        if "kpi" in log.keys():
            table += f"{(log['name'][:45] + '…' if len(log['name']) > 45 else log['name']):<50}"
            table += f"\t{log['activating'] / 1000:>}\n"
    return table


def main():
    """Main program function"""

    args = parse_cli()

    # Initialize vars
    dbus_list = []
    file_prefix = f"{args.output_directory}/"
    description = ""

    if args.post_process:
        try:
            with open(args.post_process, "r", encoding="utf-8") as input_file:
                print(f"\nPost-processing {input_file.name} ...")
                input_json = json.loads(input_file.read())
                metadata = input_json["metadata"]
                cli_offset_time_us = metadata["user_supplied_offset_us"]
                cli_offset_time_ms = cli_offset_time_us / 1000
                cntvct_offset_us = metadata["cntvct_offset_us"]
                system_init_us = cntvct_offset_us + cli_offset_time_us
                log_to_completion = metadata["log_to_completion"]
                max_time_us = metadata["log_limit_us"]
                description = metadata["description"]

                all_logs = input_json["boot_logs"]
                for log in all_logs:
                    if "systemd-dbus" in log["source"]:
                        dbus_list.append(log)

                # Get the existing file prefix from the input file
                p = re.compile("^[^-]*")
                file_prefix += p.findall(input_file.name)[0]

                artifacts_message = (
                    f"Processed existing log JSON file {input_file.name}\n"
                )

        except (FileNotFoundError, PermissionError):
            logging.error(
                "Unable to open file %s for post processing.",
                args.post_process,
            )
            sys.exit(1)

    else:
        timestamp = datetime.now()

        cntvct_offset_us = cntvct_offset()
        # Combine user specified offset with CNTVCT offset for wall clock
        cli_offset_time_ms = args.offset
        cli_offset_time_us = round(cli_offset_time_ms * 1000)
        system_init_us = cntvct_offset_us + cli_offset_time_us

        # Get systemd-analyze plot logs
        # We need to do this before dbus because if the boot up is not finished
        # as reported by systemd then dbus will have 0 for FinishTimestampMonotonic
        # which will cause a divide-by-zero error
        saplot_list = []
        saplot_retries = 10
        saplot_retry_seconds = 10
        saplot_current = 0
        while saplot_current < saplot_retries:
            saplot_exit, saplot_out, saplot_error = run_command(
                "systemd-analyze plot --json=short --no-legend"
            )
            if saplot_exit and "Bootup is not yet finished" in saplot_error:
                logging.warning(
                    "systemd-analyze reports boot up not yet finished; "
                    f"retrying in {saplot_retry_seconds} seconds ..."
                )
                saplot_current += 1
                sleep(saplot_retry_seconds)
                continue
            elif saplot_exit:
                logging.error(f"systemd-analyze plot command failed: {saplot_error}")
                sys.exit(1)
            else:
                break
        else:
            logging.error("systemd-analyze retries exceeded")
            sys.exit(1)

        # Get dbus timings for systemd
        system_bus = dbus.SystemBus()
        systemd_properties = system_bus.call_blocking(
            "org.freedesktop.systemd1",
            "/org/freedesktop/systemd1",
            "org.freedesktop.DBus.Properties",
            "GetAll",
            "s",
            [""],
        )

        # DEBUG
        # print(json.dumps(systemd_properties, indent=1))

        # Capture machine-readable hierarchy of stages

        # Note: We can't consistently get Firmware and Loader times from all platforms
        # dbus_list.append(
        #     get_time_diff(
        #         systemd_properties,
        #         "Firmware",
        #         None,
        #         "FirmwareTimestampMonotonic",
        #         "LoaderTimestampMonotonic",
        #         cntvct_offset_us,
        #     )
        # )
        # dbus_list.append(
        #     get_time_diff(
        #         systemd_properties,
        #         "Loader",
        #         None,
        #         "LoaderTimestampMonotonic",
        #         "LoaderTimestampMonotonic",
        #         cntvct_offset_us,
        #     )
        # )

        sysinit_timing = {
            "source": "systemd-dbus",
            "name": "SystemInit",
            "activating": 0,
            "activated": system_init_us,
            "time": system_init_us,
        }
        dbus_list.append(sysinit_timing)

        cli_offset_timing = {
            "source": "cli",
            "name": "KernelTimerOffset",
            "activating": 0,
            "activated": cli_offset_time_us,
            "time": cli_offset_time_us,
            "parent": "SystemInit",
        }
        dbus_list.append(cli_offset_timing)

        cntvct_offset_timing = {
            "source": "systemd-dbus",
            "name": "SystemInitCNTVCT",
            "activating": cli_offset_time_us,
            "activated": system_init_us,
            "time": cntvct_offset_us,
            "parent": "SystemInit",
        }
        dbus_list.append(cntvct_offset_timing)

        kernel_timing = get_time_diff(
            systemd_properties,
            "Kernel",
            None,
            "KernelTimestampMonotonic",
            "InitRDTimestampMonotonic" if systemd_properties['InitRDTimestampMonotonic'] else "UserspaceTimestampMonotonic",
            system_init_us,
        )
        dbus_list.append(kernel_timing)

        if systemd_properties['InitRDTimestampMonotonic']:
            initrd_timing = get_time_diff(
                systemd_properties,
                "InitRD",
                None,
                "InitRDTimestampMonotonic",
                "UserspaceTimestampMonotonic",
                system_init_us,
            )
            dbus_list.append(initrd_timing)

            initrd_systemd_init_timing = get_time_diff(
                systemd_properties,
                "InitRDSystemdInit",
                "InitRD",
                "InitRDTimestampMonotonic",
                "InitRDUnitsLoadFinishTimestampMonotonic",
                system_init_us,
            )
            dbus_list.append(initrd_systemd_init_timing)

            initrd_security_timing = get_time_diff_for_key_prefix(
                systemd_properties, "InitRDSecurity", "InitRDSystemdInit", system_init_us
            )
            dbus_list.append(initrd_security_timing)

            initrd_generators_timing = get_time_diff_for_key_prefix(
                systemd_properties, "InitRDGenerators", "InitRDSystemdInit", system_init_us
            )
            dbus_list.append(initrd_generators_timing)

            initrd_units_load_timing = get_time_diff_for_key_prefix(
                systemd_properties, "InitRDUnitsLoad", "InitRDSystemdInit", system_init_us
            )
            dbus_list.append(initrd_units_load_timing)

            initrd_units_load_finish_ts = (
                systemd_properties["InitRDUnitsLoadFinishTimestampMonotonic"]
                + system_init_us
            )
            initrd_first_service_timestamp = {
                "source": "systemd-dbus",
                "name": f"Initrd systemd units load finished at {initrd_units_load_finish_ts / 1000000} ms",
                "time": 0,
                "activating": initrd_units_load_finish_ts,
                "parent": "InitRD",
                "kpi": "True",
            }
            dbus_list.append(initrd_first_service_timestamp)

        switchroot_timing = get_time_diff(
            systemd_properties,
            "Switchroot",
            None,
            "UserspaceTimestampMonotonic",
            "FinishTimestampMonotonic",
            system_init_us,
        )
        dbus_list.append(switchroot_timing)

        systemd_init_timing = get_time_diff(
            systemd_properties,
            "SystemdInit",
            "Switchroot",
            "UserspaceTimestampMonotonic",
            "UnitsLoadFinishTimestampMonotonic",
            system_init_us,
        )
        dbus_list.append(systemd_init_timing)

        security_timing = get_time_diff_for_key_prefix(
            systemd_properties, "Security", "SystemdInit", system_init_us
        )
        dbus_list.append(security_timing)

        generators_timing = get_time_diff_for_key_prefix(
            systemd_properties,
            "Generators",
            "SystemdInit",
            system_init_us,
        )
        dbus_list.append(generators_timing)

        units_load_timing = get_time_diff_for_key_prefix(
            systemd_properties, "UnitsLoad", "SystemdInit", system_init_us
        )
        dbus_list.append(units_load_timing)

        root_units_load_finish_ts = (
            systemd_properties["UnitsLoadFinishTimestampMonotonic"] + system_init_us
        )
        first_service_timestamp = {
            "source": "systemd-dbus",
            "name": f"Root systemd units load finished at",
            "time": root_units_load_finish_ts,
            "activating": root_units_load_finish_ts,
            "parent": "Switchroot",
            "kpi": "True",
        }
        dbus_list.append(first_service_timestamp)

        overall_timing = {
            "source": "systemd-dbus",
            "name": "Overall",
            "time": systemd_properties["FinishTimestampMonotonic"] + system_init_us,
            "activating": 0,
            "ignore": "True",
        }
        dbus_list.append(overall_timing)

        dbus_list.append(
            {
                "source": "systemd-dbus",
                "name": "NInstalledJobs",
                "count": systemd_properties["NInstalledJobs"],
                "time": 0,
                "activating": 0,
                "ignore": "True",
            }
        )

        dbus_list.append(
            {
                "source": "systemd-dbus",
                "name": "NFailedJobs",
                "count": systemd_properties["NFailedJobs"],
                "time": 0,
                "activating": 0,
                "ignore": "True",
            }
        )

        # Set the max time used for filtering other logs
        log_to_completion = False
        max_time_us = systemd_properties["FinishTimestampMonotonic"] + system_init_us
        try:
            max_time_us = float(args.max_time * 1000)
        except ValueError:
            log_to_completion = True

        # Get journalctl logs
        journalctl_exit, journalctl_out, journalctl_err = run_command(
            "journalctl -b0 -o short-delta --no-pager"
        )

        if journalctl_exit:
            logging.error("journalctl command failed: %s", journalctl_err)
            sys.exit(1)

        # Parse journalctl log messages and timings
        journalctl_lines = journalctl_out.splitlines()

        journalctl_list = []

        for line in journalctl_lines:
            log = re.split("] | <", line, maxsplit=2)
            try:
                journalctl_dict = {}
                journalctl_dict["source"] = "journalctl"
                hostname_filtered = " ".join(log[2].split()[1:])
                journalctl_dict["name"] = hostname_filtered
                journalctl_dict["activating"] = int(float(log[0].strip("[ ")) * 1000000)
                journalctl_dict["time"] = int(float(log[1].strip(">")) * 1000000)
                filter_check = log_filter(
                    journalctl_dict, max_time_us, args.kpi_re_pattern, system_init_us
                )
                if filter_check:
                    journalctl_list.append(filter_check)
            except IndexError:
                continue

        # DEBUG
        # print(journalctl_list)
        # Parse systemd-analyze plot log messages and timings
        saplot_dict = ast.literal_eval(saplot_out)
        for log in saplot_dict:
            log["source"] = "saplot"
            filter_check = log_filter(
                log, max_time_us, args.kpi_re_pattern, system_init_us
            )
            if filter_check:
                saplot_list.append(filter_check)

        # DEBUG
        # print(saplot_list)

        if args.description:
            description = args.description

        metadata = {}
        metadata["hostname"] = socket.gethostname()
        metadata["date"] = timestamp.isoformat()
        metadata["platform"] = platform.platform(terse=True)
        metadata["kernel"] = platform.release()
        metadata["model"] = platform.processor()
        metadata["architecture"] = platform.machine()
        metadata["system"] = platform.system()
        metadata["description"] = description
        metadata["cntvct_offset_us"] = cntvct_offset_us
        metadata["user_supplied_offset_us"] = cli_offset_time_us
        metadata["wall_clock_offset_us"] = system_init_us
        metadata["log_to_completion"] = log_to_completion
        metadata["log_limit_us"] = max_time_us

        # DEBUG
        # print(json.dumps(metadata, indent=2))

        # Merge logs
        all_logs = dbus_list + journalctl_list + saplot_list

        logs_dict = {
            "metadata": metadata,
            "boot_logs": all_logs,
        }

        file_prefix += (
            f"{metadata['hostname']}_{timestamp.strftime('%m_%d_%Y_%H_%M_%S').strip()}"
        )

        # Output the log data as JSON
        log_output_file = f"{file_prefix}-boot_time_logs.json"
        with io.open(log_output_file, "w", encoding="utf8") as out_file:
            logs_string = json.dumps(
                logs_dict,
                indent=4,
                sort_keys=False,
                separators=(",", ": "),
                ensure_ascii=False,
            )
            out_file.write(f"{logs_string}\n")
            artifacts_message = f"Created log JSON file at {log_output_file}\n"

    # Remove log items marked to ignore
    cleaned_logs = []
    for log in all_logs:
        if "ignore" not in log.keys():
            cleaned_logs.append(log)

    # DEBUG
    # print(json.dumps(cleaned_logs, indent=1))

    # Build and export the log chart
    if chart:
        artifacts_message += chart.build(
            cleaned_logs,
            metadata,
            file_prefix,
        )
        artifacts_message += "\n"

    # Return report to stdout
    offset_output_message = "\nNOTES: \n======\n"
    if not systemd_properties['InitRDTimestampMonotonic']:
        offset_output_message += (
            "- System does not appear to have an initrd/initramfs.\n"
        )

    if cli_offset_time_ms != 0:
        offset_output_message += (
            "- A user-specified offset of "
            + str(cli_offset_time_ms)
            + " ms have been applied to SystemInit.\n"
        )

    if cntvct_offset_us:
        offset_output_message += (
            "- CNTVCT offset of "
            + str(cntvct_offset_us / 1000)
            + " ms has been applied to SystemInit.\n"
        )

    if system_init_us:
        offset_output_message += (
            "- Offset(s) applied. Times will be reported with an offset of "
            + str(system_init_us / 1000)
            + " ms.\n"
        )
    else:
        offset_output_message += (
            "- No offsets found. Times will be reported in kernel time.\n"
        )

    if log_to_completion:
        offset_output_message += (
            "- Default to logging until completion of switch root.\n"
        )

    offset_output_message += f"- Logs limit: {(max_time_us / 1000)} ms\n"

    print(offset_output_message)

    # DEBUG
    # print(json.dumps(systemd_properties, indent=1))
    # print(json.dumps(dbus_list, indent=1))

    # Friendly print systemd stages from list
    print("\n# System boot time analysis\n")
    systemd_stages_output = print_systemd_stages(dbus_list)
    print(systemd_stages_output)

    # Print a table of KPIs and timestamps
    table_output = kpi_table(cleaned_logs)
    print(table_output)

    # Build satime, initramfs, and switchroot dictionaries from boot logs
    satime_dict = {}
    initramfs_dict = {}
    switchroot_dict = {}
    for log in dbus_list:
        if log["source"] == "systemd-dbus":
            if log["name"] == "SystemInit":
                satime_dict["sysinit"] = log["time"] / 1000000
                continue
            if log["name"] == "Kernel":
                satime_dict["kernel"] = log["time"] / 1000000
                continue
            if log["name"] == "InitRD":
                satime_dict["initrd"] = log["time"] / 1000000
                continue
            if log["name"] == "Switchroot":
                satime_dict["switchroot"] = log["time"] / 1000000
                continue
            if log["name"] == "Overall":
                satime_dict["total"] = log["time"] / 1000000
                continue
            if log["name"] == "InitRDSystemdInit":
                initramfs_dict["systemd_init"] = log["time"] / 1000000
                continue
            if log["name"] == "InitRDSecurity":
                initramfs_dict["security"] = log["time"] / 1000000
                continue
            if log["name"] == "InitRDGenerators":
                initramfs_dict["generators"] = log["time"] / 1000000
                continue
            if log["name"] == "InitRDUnitsLoad":
                initramfs_dict["units_load"] = log["time"] / 1000000
                continue
            if "First Initrd Service Started At" in log["name"]:
                initramfs_dict["first_service_ts"] = log["activating"] / 1000000
                continue
            if log["name"] == "SystemdInit":
                switchroot_dict["systemd_init"] = log["time"] / 1000000
                continue
            if log["name"] == "Security":
                switchroot_dict["security"] = log["time"] / 1000000
                continue
            if log["name"] == "Generators":
                switchroot_dict["generators"] = log["time"] / 1000000
                continue
            if log["name"] == "UnitsLoad":
                switchroot_dict["units_load"] = log["time"] / 1000000
                continue
            if "First Root Service Started At" in log["name"]:
                switchroot_dict["first_service_ts"] = log["activating"] / 1000000
                continue

    # Merge summary item
    summary_dict = {}
    summary_dict["system_config"] = metadata
    summary_dict["satime"] = satime_dict
    summary_dict["initramfs"] = initramfs_dict
    summary_dict["switchroot"] = switchroot_dict
    summary_dict["system_config"]["description"] = description

    # DEBUG
    # print(json.dumps(summary_dict, indent=1))

    # Output the summary data as JSON
    summary_output_file = f"{file_prefix}-boot_time_summary.json"
    with io.open(summary_output_file, "w", encoding="utf8") as out_file:
        summary_string = json.dumps(
            summary_dict,
            indent=4,
            sort_keys=False,
            separators=(",", ": "),
            ensure_ascii=False,
        )
        out_file.write(f"{summary_string}\n")
        artifacts_message += f"Created summary JSON file at {summary_output_file}\n"

    log_output_file = f"{file_prefix}-boot_time_output.log"
    with io.open(log_output_file, "w", encoding="utf8") as out_file:
        # Write command used to file
        if description:
            out_file.write(f"Description: {description}\n\n")
        out_file.write(f"Command: {' '.join(sys.argv)}\n\n")
        out_file.write(f"{offset_output_message}\n")
        out_file.write(f"{systemd_stages_output}\n")
        out_file.write(f"{table_output}\n")
        artifacts_message += f"Created output log file at {log_output_file}\n"

    print("\n## Artifacts")
    print(artifacts_message)


if __name__ == "__main__":
    main()
